/*
Copyright (C) 2002 Chr. Clemens Lee <clemens@kclee.com>.

This file is part of JIM - Java Information Manager
(http://www.kclee.com/clemens/java/jim/).

Jim is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

Jim is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Jim; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

package jim;

import ccl.util.Util;
import java.util.Vector;
import javax.swing.tree.DefaultMutableTreeNode;
import org.xml.sax.AttributeList;
import org.xml.sax.HandlerBase;

/**
 * Sax DocumentHandler for Jim.
 *
 * @author  Chr. Clemens Lee
 * @version $Id: JimDocumentHandler.java,v 1.3 2002/04/21 13:25:21 clemens Exp clemens $
 */
public class JimDocumentHandler extends HandlerBase {
    private static final String S_NOTE        = "NOTE";
    private static final String S_KEY         = "KEY";
    private static final String S_VALUE       = "VALUE";

    private static final String S_PROJECT     = "PROJECT";
    private static final String S_NAME        = "NAME";
    private static final String S_TODO        = "TODO";
    private static final String S_ID          = "ID";
    private static final String S_TITLE       = "TITLE";
    private static final String S_COMMENT     = "COMMENT";
    private static final String S_START       = "START";
    private static final String S_CLOSED      = "CLOSED";
    private static final String S_FINISHED    = "FINISHED";
    private static final String S_DEADLINE    = "DEADLINE";
    private static final String S_PRIORITY    = "PRIORITY";
    private static final String S_DEPENDSON   = "DEPENDSON";
    private static final String S_BLOCKS      = "BLOCKS";

    private static final String S_ANNIVERSARY = "ANNIVERSARY";
    private static final String S_DATE        = "DATE";

    private static final int    NONE        = -1;
    private static final int    KEY         =  0;
    private static final int    VALUE       =  1;
    private static final int    PROJECT     =  2;
    private static final int    ANNIVERSARY =  3;

    private int _status = NONE;

    private String _sTimezoneID = null;

    // notes
    private DefaultMutableTreeNode _tnNotesRoot      = null;
    private DefaultMutableTreeNode _tnCurrent        = null;
    private String                 _sKey             = null;
    private StringBuffer           _sbValue          = null;
    private StringBuffer           _sbText           = null;

    // todos
    private DefaultMutableTreeNode _tnRootProject    = null;
    private DefaultMutableTreeNode _tnCurrentProject = null;
    private ToDoTaskVector         _vToDoTasks       = null;
    private ToDoTask               _pToDoTask        = null;

    // anniversaries
    private Vector                 _vAnniversaries   = null;
    private Anniversary            _pAnniversary     = null;

    public JimDocumentHandler( String sTimezoneID_ ) {
        super();
        
        _sTimezoneID    = sTimezoneID_;
        _status         = NONE;
        _vAnniversaries = new Vector();
        _sbValue = new StringBuffer( 1000 );
        _sbText  = new StringBuffer( 1000 );
    }
    
    public void startElement( String        name_, 
                              AttributeList attributes_ )
    {
        if ( name_.equals( S_KEY ) ) {
            _status = KEY;
            _sKey = "";
        } else if ( name_.equals( S_VALUE ) ) {
            _status = VALUE;
            _sbValue.setLength( 0 );
        } else if ( name_.equals( S_TODO ) ) {
            _pToDoTask = new ToDoTask( _sTimezoneID );
        } else if ( name_.equals( S_PROJECT ) ) {
            _status = PROJECT;
        } else if ( name_.equals( S_ANNIVERSARY ) ) {
            _status = ANNIVERSARY;
        }
        _sbText.setLength( 0 );
    }

    public void endElement( String name_ ) {
        if ( name_.equals( S_VALUE ) ) {
            // time to create a note and node
            Note                   note  = new Note( _sKey, _sbValue.toString() );
            DefaultMutableTreeNode tnNew = new DefaultMutableTreeNode( note );
            if ( _tnNotesRoot == null ) {
                _tnNotesRoot = tnNew;
            }
            if ( _tnCurrent != null ) {
                _tnCurrent.add( tnNew );
            }
            _tnCurrent = tnNew;
        } else if ( name_.equals( S_NOTE ) ) {
            _tnCurrent = (DefaultMutableTreeNode)_tnCurrent.getParent();
        } else if ( _status == PROJECT && name_.equals( S_NAME ) ) {
            // this is actually the beginning of a project
            _vToDoTasks = new ToDoTaskVector( _sbText.toString() );
            DefaultMutableTreeNode tnNew = new DefaultMutableTreeNode
                ( _vToDoTasks );
            if ( _tnRootProject == null ) {
                _tnRootProject = tnNew;
            }
            if ( _tnCurrentProject != null ) {
                _tnCurrentProject.add( tnNew );
            }
            _tnCurrentProject = tnNew;
        } else if ( name_.equals( S_PROJECT ) ) {
            _tnCurrentProject = (DefaultMutableTreeNode)_tnCurrentProject.getParent();
        } else if ( name_.equals( S_ID ) ) {
            _pToDoTask.setId( Util.atol( _sbText.toString() ) );
        } else if ( name_.equals( S_TITLE ) ) {
            _pToDoTask.setTitle( _sbText.toString() );
        } else if ( name_.equals( S_COMMENT ) ) {
            _pToDoTask.setComment( _sbText.toString() );
        } else if ( name_.equals( S_START ) ) {
            _pToDoTask.setStart( _sbText.toString() );
        } else if ( name_.equals( S_CLOSED ) ) {
            _pToDoTask.setClosed( _sbText.toString() );
        } else if ( name_.equals( S_FINISHED ) ) {
            _pToDoTask.setFinished( _sbText.toString() );
        } else if ( name_.equals( S_DEADLINE ) ) {
            _pToDoTask.setDeadline( _sbText.toString() );
        } else if ( name_.equals( S_DEPENDSON ) ) {
            _pToDoTask.addDependency( _sbText.toString() );
        } else if ( name_.equals( S_BLOCKS ) ) {
            _pToDoTask.addBlock( _sbText.toString() );
        } else if ( name_.equals( S_PRIORITY ) ) {
            _pToDoTask.setPriority( _sbText.toString() );

            // we now know it must be the end of the todo task
            _vToDoTasks.addElement( _pToDoTask );
        } else if ( _status == ANNIVERSARY && name_.equals( S_NAME ) ) {
            _pAnniversary = new Anniversary();
            _pAnniversary.setName( _sbText.toString() );
        } else if ( name_.equals( S_DATE ) ) {
            _pAnniversary.setDate( _sbText.toString() );
            _vAnniversaries.addElement( _pAnniversary );
        }

        _status = NONE;
    }

    public void characters( char ac_[], int start_, int length_ ) {
        if ( _status == KEY ) {
            _sKey += new String( ac_, start_, length_ );
        } else if ( _status == VALUE ) {
            _sbValue.append( ac_, start_, length_ );
        } else {
            _sbText .append( ac_, start_, length_ );
        }
    }

    public DefaultMutableTreeNode getNotes() {
        return _tnNotesRoot;
    }

    public DefaultMutableTreeNode getToDos() {
        return _tnRootProject;
    }

    public Vector                 getAnniversaries() {
        return _vAnniversaries;
    }
}
