/*
Copyright (C) 2002 Chr. Clemens Lee <clemens@kclee.com>.

This file is part of JIM - Java Information Manager
(http://www.kclee.com/clemens/java/jim/).

Jim is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

Jim is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Jim; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

package jim;

import ccl.util.Util;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Vector;

/**
 * 
 * @author  Chr. Clemens Lee
 * @version $Id: ToDoTask.java,v 1.5 2002/04/21 13:25:25 clemens Exp clemens $
 */
public class ToDoTask implements Cloneable {
    public static Priority A = new Priority( "A", 1.0f );
    public static Priority B = new Priority( "B", 2.0f );
    public static Priority C = new Priority( "C", 3.0f );
    public static Priority D = new Priority( "D", 4.0f );

    private String _sTitle   = null;
    private String _sComment = null;
    private Date   _dtStart  = null;

    /** 
     * Each task should have a distinct ID. In practice
     * we want to use the current time in millis.
     */
    private long   _lId      = 0;

    /**
     * No need to work on it earlier. If null, work could or
     * should start immediately.
     */
    private Date _dtWorkStart = null;

    /**
     * Should be finished at this day. If null, it should be
     * finished as soon as possible or when you find the time
     * (or take as much time as you want :-).
     */
    private Date _dtDeadline = null;

    /**
     * Either _dtFinished or _dtClosed have to be set to either
     * make clear this job is done or will nether be done because
     * it is not necessary to finish.
     */
    private Date _dtFinished = null;
	
    /**
     * Either _dtFinished or _dtClosed have to be set to either
     * make clear this job is done or will nether be done because
     * it is not necessary to finish.
     */
    private Date _dtClosed = null;

    /**
     * 0 means, your life depends on it. Top priority is 1.0.
     * Lowest priority should be 3.0 - 5.0, haven't made up
     * my mind yet.
     */
    private Priority _pPriority = null;

    private String _sTimezoneID = null;

    /**
     * This vector contains id strings to other tasks which
     * have to be finished before this one can be finished.
     */
    private Vector _vDependencies = new Vector();

    /**
     * This vector contains id strings to other tasks which
     * can't be finished unless this task is finished.
     */
    private Vector _vBlocks       = new Vector();

    public void setDependencies( Vector vDependencies_ ) {
        _vDependencies = (Vector)vDependencies_.clone();
    }

    public Vector getDependencies() {
        return (Vector)_vDependencies.clone();
    }

    public void setBlocks( Vector vBlocks_ ) {
        _vBlocks = (Vector)vBlocks_.clone();
    }

    public Vector getBlocks() {
        return (Vector)_vBlocks.clone();
    }

    public void setDependencies( String sDependencies_ ) {
        _vDependencies = Util.stringToLines( Util.replace( sDependencies_, " ", "" ), ',' );
    }

    public void setBlocks( String sBlocks_ ) {
        _vBlocks = Util.stringToLines( Util.replace( sBlocks_, " ", "" ), ',' );
    }

    public String getDependenciesString() {
        return Util.concat( _vDependencies, ", " );
    }

    public String getBlocksString() {
        return Util.concat( _vBlocks, ", " );
    }

    public void addDependency( String sId_ ) {
        _vDependencies.addElement( sId_ );
    }

    public void addBlock( String sId_ ) {
        _vBlocks.addElement( sId_ );
    }

    static private long _lLastId = 0;

    static private synchronized long _getNextId() {
        long lNextId = _lLastId + 1;

        _lLastId = lNextId;

        return lNextId;
    }

    public ToDoTask() {
        super();

        _dtStart     = new Date();
        _sTitle      = "";
        _sComment    = "";
        _pPriority   = C;
        _sTimezoneID = null;
        _lId         = _getNextId();
    }

    public ToDoTask( String sTimezoneID_ ) {
        this();

        _sTimezoneID = sTimezoneID_;
    }

    public void setPriority( Priority pPriority_ ) {
        _pPriority = pPriority_;
    }

    public void setPriority( String sPriority_ ) {
        if ( sPriority_ == null ) {
            _pPriority = C;
        } else if ( sPriority_.equals( "A" ) ) {
            _pPriority = A;
        } else if ( sPriority_.equals( "B" ) ) {
            _pPriority = B;
        } else if ( sPriority_.equals( "D" ) ) {
            _pPriority = D;
        } else {
            _pPriority = C;
        }
    }

    public Priority getPriority() {
        return _pPriority;
    }

    /**
     * The ToDo task is neither finished nor closed.
     */
    public boolean isOpen() {
        return ((_dtFinished == null) &&
                (_dtClosed == null));
    }

    public String getStart() {
        return Util.getStandardDate( _dtStart );
    }

    /**
     * For example: 1999-11-26
     *
     * @return   true on parse error, still current date is set.
     */
    public boolean setStart( String sStart_ ) {
        try {
            String sYear = sStart_.substring( 0, 4 );
            String sMonth = sStart_.substring( 5, 7 );
            String sDay = sStart_.substring( 8, 10 );
            int year = Util.atoi( sYear );
            int month = Util.atoi( sMonth ) - 1;
            int day = Util.atoi( sDay );
            Util.debug( this, "setStart(..).year: " + year );
            Util.debug( this, "setStart(..).month: " + month );
            Util.debug( this, "setStart(..).day: " + day );
            Calendar pCalendar = new GregorianCalendar();
            pCalendar.setTime( _dtStart );
            pCalendar.set( Calendar.YEAR, year );
            pCalendar.set( Calendar.MONTH, month );
            pCalendar.set( Calendar.DAY_OF_MONTH, day );
            _dtStart.setTime( pCalendar.getTime().getTime() );
            Util.debug( this, "setStart(..)._dtStart: " + _dtStart );
        } catch( Exception pException ) {
            Util.debug( this, "setStart(..).pException: " + pException );
            _dtStart = new Date();
            
            return true;
        }
        Util.panicIf( !sStart_.equals( getStart() ),
                      "_dtStart: " + _dtStart );

        return false;
    }

    public void setTitle( String sTitle_ ) {
        _sTitle = sTitle_;
    }

    /**
     * Returns 0 for green, 1 for yellow, 2 for red,
     * 3 for finished, 4 for closed.
     */
    public int getStatus() {
        if ( isFinished() ) {
            return 3;
        }
        if ( isClosed() ) {
            return 4;
        }

        if ( !Util.isEmpty( getDeadline() ) ) {
            String sToday = Util.getDate
                   ( Util.getCalendar( _sTimezoneID ) );
            if ( sToday.compareTo( getDeadline() ) <= 0 ) {
                return 1;
            } else {
                return 2;
            }
        }

        return 0;
    }

    private static final String[] AS_STATUS_HTML = 
    {
        "<font color=\"00FF00\">Green</font>",
        "<font color=\"00FFFF\">Yellow</font>",
        "<font color=\"FF0000\">Red</font>",
        "<font color=\"0000FF\">Done</font>",
        "<font color=\"000000\">Closed</font>"
    };
        
    public String getStatusHTML() {
        return AS_STATUS_HTML[ getStatus() ];
    }

    public String getDate() {
        int status = getStatus();
        if ( status == 3 ) {
            return getFinished();
        } else if ( status == 4 ) {
            return getClosed();
        } else if ( status == 0 ) {
            return getStart();
        }

        // yellow or red
        return getDeadline();
    }

    public String toString() {
        String sRetVal = getPriority() + "   ";
        if ( isFinished() ) {
            sRetVal += "+   " + getFinished() + "   ";
        }
        if ( isClosed() ) {
            sRetVal += "-   " + getClosed() + "   ";
        }
        if ( isOpen() ) {
            if ( !Util.isEmpty( getDeadline() ) ) {
                String sToday = Util.getStandardDate( new Date() );
                if ( sToday.compareTo( getDeadline() ) <= 0 ) {
                    sRetVal += "Y   " + getDeadline() + "   ";
                } else {
                    sRetVal += "R   " + getDeadline() + "   ";
                }
            } else {
                sRetVal += "G   " + getStart() + "   ";
            }
        }
        sRetVal += getTitle();

        return sRetVal;
    }

    public String getTitle() {
        return _sTitle;
    }

    public String getComment() {
        return _sComment;
    }

    public void setComment( String sComment_ ) {
        Util.panicIf( sComment_ == null );

        _sComment = sComment_;
    }

    public boolean isFinished() {
        if ( Util.isEmpty( getFinished() ) ) {
            return false;
        }

        return true;
    }

    public boolean isClosed() {
        return !Util.isEmpty( getClosed() );
    }
    
    public void setClosed( String sClosed_ ) {
        _dtClosed = Util.stringToDate( sClosed_ );
    }

    public void setFinished( String sFinished_ ) {
        _dtFinished = Util.stringToDate( sFinished_ );
    }

    public void setDeadline( String sDeadline_ ) {
        _dtDeadline = Util.stringToDate( sDeadline_ );
    }

    /**
     * @return   "" or well defined string.
     */
    public String getDeadline() {
        String sDeadline = "";
        if ( _dtDeadline != null ) {
            sDeadline = Util.getStandardDate( _dtDeadline );
        }

        return sDeadline;
    }

    /**
     * @return   "" or well defined string.
     */
    public String getClosed() {
        String sClosed = "";
        if ( _dtClosed != null ) {
            sClosed = Util.getStandardDate( _dtClosed );
        }

        return sClosed;
    }

    /**
     * @return   "" or well defined string.
     */
    public String getFinished() {
        String sFinished = "";
        if ( _dtFinished != null ) {
            sFinished = Util.getStandardDate( _dtFinished );
        }

        return sFinished;
    }

    /**
     * End date is either closed date, finished date, or null.
     */
    public String getEnd() {
        String sRetVal = getClosed();
        if ( Util.isEmpty( sRetVal ) ) {
            sRetVal = getFinished();
        }

        return sRetVal;
    }

    public long getId() {
        return _lId;
    }

    public void setId( long lId_ ) {
        _lId = lId_;

        if ( _lId > _lLastId ) {
	         _lLastId = _lId;
        }
    }

    static public void resetLastId() {
        _lLastId = 0;
    }

    /**
     * Warning, this returns only true if two tasks really are
     * the same, not if they have equal values.
     */
    public boolean equals( Object oToDoTask ) {
        return( this == oToDoTask );
    }

    public Object clone() {
        ToDoTask pToDoTask = new ToDoTask( _sTimezoneID );
        pToDoTask._sTitle = this._sTitle;
        pToDoTask._sComment = this._sComment;

        pToDoTask._dtStart = null;
        if ( this._dtStart != null ) {
            pToDoTask._dtStart = new Date();
            pToDoTask._dtStart.setTime( this._dtStart.getTime() );
        }

        pToDoTask._dtWorkStart = null;
        if ( this._dtWorkStart != null ) {
            pToDoTask._dtWorkStart = new Date();
            pToDoTask._dtWorkStart.setTime( this._dtWorkStart.getTime() );
        }

        pToDoTask._dtDeadline = null;
        if ( this._dtDeadline != null ) {
            pToDoTask._dtDeadline = new Date();
            pToDoTask._dtDeadline.setTime( this._dtDeadline.getTime() );
        }

        pToDoTask._dtFinished = null;
        if ( this._dtFinished != null ) {
            pToDoTask._dtFinished = new Date();
            pToDoTask._dtFinished.setTime( this._dtFinished.getTime() );
        }

        pToDoTask._dtClosed = null;
        if ( this._dtClosed != null ) {
            pToDoTask._dtClosed = new Date();
            pToDoTask._dtClosed.setTime( this._dtClosed.getTime() );
        }

        pToDoTask._pPriority = this._pPriority;

        pToDoTask._lId       = this._lId;

        return pToDoTask;
    }

    public static class Priority extends Object {
        private String _sName = null;
        private String _sComment = null;
        private float _fPriority = 3.0f;

        private Priority( String sName_, float fPriority_ ) {
            super();
            
            _sName = sName_;
            _fPriority = fPriority_;
        }

        public String toString() {
            return _sName;
        }

        public int compareTo( Priority pPriority_ ) {
            return toString().compareTo( pPriority_.toString() );
        }
    }
}
