/*
Copyright (C) 2002 Chr. Clemens Lee <clemens@kclee.com>.

This file is part of JIM - Java Information Manager
(http://www.kclee.com/clemens/java/jim/).

Jim is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

Jim is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Jim; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

package jim;

import ccl.swing.ComponentHierachyFrame;
import ccl.swing.MainJFrame;
import ccl.swing.MainJMenuBar;
import ccl.swing.RunnableAction;
import ccl.swing.SplashWindow;
import ccl.swing.SwingUtil;
import ccl.util.FileUtil;
import ccl.util.Init;
import ccl.util.Util;
import com.sun.xml.parser.Resolver;
import com.sun.xml.tree.ElementNode;
import com.sun.xml.tree.TreeWalker;
import com.sun.xml.tree.XmlDocument;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.Frame;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.URL;
import java.util.Calendar;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.text.JTextComponent;
import javax.swing.text.Keymap;
import javax.swing.text.TextAction;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeNode;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.xml.sax.InputSource;
import org.xml.sax.Parser;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.ParserFactory;

/**
 *
 *
 * @author Chr. Clemens Lee
 * @version $Id: ViewController.java,v 1.13 2002/04/21 13:49:08 clemens Exp clemens $
 */
public class ViewController extends MainJFrame
                            implements ActionListener
{
    private static final String S_PARSER          = "com.sun.xml.parser.Parser";

    private static final String S_MF_NEW          = "&New";
    private static final String S_MF_OPEN         = "&Open...";
    private static final String S_MF_SAVE         = "&Save";
    private static final String S_MF_SAVE_AS      = "Save &As...";
    private static final String S_MF_EXPORT_NOTES = "Export Notes...";
    private static final String S_MF_EXPORT_NOTE  = "Export Note...";
    private static final String S_MF_EXPORT_TODO  = "Export ToDo List(s)...";

    private static final String S_NEXTRECALL = "nr";

    private AnniversariesPanel _pAnniversariesPanel = null;
    private NotesPanel         _pNotesPanel         = null;
    private ToDoPanel          _pToDoPanel          = null;
    private Rectangle          _bounds              = null;

    private boolean _bBatchMode = false;

    protected boolean _confirmClose() {
        if ( _bInitialize ) 
        {
            return true;
        }

        if ( _pToDoPanel.isTaskVisible() ) 
        {
            SwingUtil.showMessage
                   ( this,
                     "You are currently editing a ToDo task.\n"
                     + "Please first finish editing this task before\n"
                     + "we will close the current project. Thanks." );

            return false;
        }

        return super._confirmClose();
    }

    private String _escapeFileName( String sFileName )
    {
        sFileName = Util.replace( sFileName, " ", "_" );
        sFileName = Util.replace( sFileName, "/", "_" );
        sFileName = Util.replace( sFileName, "\u00dc", "Ue" );
        sFileName = Util.replace( sFileName, "\u00fc", "ue" );
        sFileName = Util.replace( sFileName, "\u00df", "ss" );
        sFileName = Util.replace( sFileName, "\u00c4", "Ae" );
        sFileName = Util.replace( sFileName, "\u00e4", "ae" );
        sFileName = Util.replace( sFileName, "\u00d6", "Oe" );
        sFileName = Util.replace( sFileName, "\u00f6", "oe" );

        return sFileName;
    }

    public void newFile() {
        if ( !_confirmClose() ) {
            return;
        }

        DefaultMutableTreeNode tnRoot = NotesPanel.createRootNode();
        _pNotesPanel.setNotes( tnRoot );

        _pToDoPanel.setToDoTasks( new ToDoTaskVector( ToDoPanel.getToDoName() ) );

        _pAnniversariesPanel.setAnniversaries( new Vector() );

        super.newFile();
    }

    public class NewAction extends RunnableAction {
        public NewAction() {
            super( "&New" );
        }

        public void run() {
            newFile();
        }
    }

    /**
     * Interactively open new file/project.
     */
    public void open() {
        if ( !_confirmClose() ) {
            return;
        }

        String sProjectFullName = null;
        Init pInit = getInit();
        String sProjectSuffix = pInit.getKeyValue( Init.getProjectSuffixKey() );
        if ( Util.isEmpty( sProjectSuffix ) ) {
            sProjectSuffix = ".jiml";
            pInit.setKeyValue( Init.getProjectSuffixKey(), sProjectSuffix );
        }
        sProjectSuffix = sProjectSuffix.substring( 1 );
        String[] asProjectSuffixes = { sProjectSuffix, 
                                       S_NEXTRECALL };
        String sOldProjectPath = pInit.getFilePath().replace('/', File.separatorChar);
        sProjectFullName = SwingUtil.getFileName
               ( this, 
                 "Choose a Project", 
                 asProjectSuffixes,
                 sOldProjectPath );
        if (sProjectFullName == null) {
            return;
        }
        if ( !FileUtil.existsFile( sProjectFullName ) ) {
            SwingUtil.showMessage( this, "Sorry, the specified file\n\n'" +
                                   sProjectFullName +
                                   "'\n\ndoes not exist!" );

            return;
        }
        // filename must end with .jiml or .nr
        if ( !Util.endsWith( sProjectFullName, sProjectSuffix ) &&
             !Util.endsWith( sProjectFullName, "." + S_NEXTRECALL ) ) 
        {
            SwingUtil.showMessage( this, "Warning: Project file name must end with '" +
                                   sProjectSuffix + "'!" );

            return;
        }
        
        open( sProjectFullName );
    }

    /**
     * Node is to be known to have only text as children
     */
    private String _getText( Node pNode_ ) {
        NodeList pNodeList = pNode_.getChildNodes();

        String sRetVal = "";
        
        for( int child = 0; child < pNodeList.getLength(); child++ ) {
            Node nNext = pNodeList.item( child );
            if ( nNext != null && nNext instanceof Text ) {
                sRetVal += ((Text)nNext).getData();
            }
        }

        return sRetVal;
    }

    private DefaultMutableTreeNode _getTreeNode( ElementNode enNote_ ) {
        TreeWalker pTreeWalker = new TreeWalker( enNote_ );

        // get key
        Node nKey = pTreeWalker.getNextElement( "KEY" );
        
        String sKey = _getText( nKey );

        // get value
        Node nValue = pTreeWalker.getNextElement( "VALUE" );

        String sValue = _getText( nValue );

        // create note
        Note pNote = new Note( sKey, sValue );

        // create tree node
        DefaultMutableTreeNode tnRetVal = new DefaultMutableTreeNode( pNote );

        // sub notes
        Node nSibling = nValue;
        while ( true ) {
            nSibling = nSibling.getNextSibling();
            if ( nSibling == null ) {
                break;
            }
            if ( !nSibling.getNodeName().equals( "NOTE" ) ) {
                continue;
            }
            DefaultMutableTreeNode tnChild = _getTreeNode( (ElementNode)nSibling );
            tnRetVal.add( tnChild );
        }

        return tnRetVal;
    }

    private DefaultMutableTreeNode _getProjectTreeNode( ElementNode enProject_ ) {
        TreeWalker pTreeWalker = new TreeWalker( enProject_ );

        // get key
        Node nName = pTreeWalker.getNextElement( "NAME" );
        
        String sName = _getText( nName );
        Util.debug( this, "_getProjectTreeNode(..).sName: " + sName );

        Node nSibling = nName;
        ToDoTaskVector vToDoTasks = new ToDoTaskVector( sName );
        while ( true ) {
            nSibling = nSibling.getNextSibling();
            Util.debug( this, "_getProjectTreeNode(..).nSibling: " +
                        nSibling );
            if ( nSibling == null ) {
                break;
            }
            if ( !nSibling.getNodeName().equals( "TODO" ) ) {
                continue;
            }
            ElementNode enToDo = (ElementNode)pTreeWalker.getNextElement( "TODO" );
            Util.panicIf( enToDo == null, "There was already a todo node, why is none found now?" );

            ElementNode enTitle = (ElementNode)pTreeWalker.getNextElement( "TITLE" );
            String sTitle = _getText( enTitle );
            ElementNode enComment = (ElementNode)pTreeWalker.getNextElement( "COMMENT" );
            String sComment = _getText( enComment );
            ElementNode enStart = (ElementNode)pTreeWalker.getNextElement( "START" );
            String sStart = _getText( enStart );
            ElementNode enClosed = (ElementNode)pTreeWalker.getNextElement( "CLOSED" );
            String sClosed = _getText( enClosed );
            ElementNode enFinished = (ElementNode)pTreeWalker.getNextElement( "FINISHED" );
            String sFinished = _getText( enFinished );
            ElementNode enDeadline = (ElementNode)pTreeWalker.getNextElement( "DEADLINE" );
            String sDeadline = _getText( enDeadline );
            ElementNode enPriority = (ElementNode)pTreeWalker.getNextElement( "PRIORITY" );
            String sPriority = _getText( enPriority );
            
            ToDoTask pToDoTask = new ToDoTask( getInit().getKeyValue( "TimezoneID" ) );
            pToDoTask.setTitle( sTitle );
            pToDoTask.setComment( sComment );
            pToDoTask.setStart( sStart );
            pToDoTask.setClosed( sClosed );
            pToDoTask.setFinished( sFinished );
            pToDoTask.setDeadline( sDeadline );
            pToDoTask.setPriority( sPriority );
            vToDoTasks.addElement( pToDoTask );
        }

        // create tree node
        DefaultMutableTreeNode tnRetVal = new DefaultMutableTreeNode( vToDoTasks );

        // sub projects
        nSibling = nName;
        while ( nSibling != null ) {
            nSibling = nSibling.getNextSibling();
            if ( nSibling == null ) {
                break;
            }
            if ( !nSibling.getNodeName().equals( "PROJECT" ) ) {
                continue;
            }
            DefaultMutableTreeNode tnChild = _getProjectTreeNode( (ElementNode)nSibling );
            tnRetVal.add( tnChild );
        }

        return tnRetVal;
    }

    private String _getTimezoneID() {
        return getInit().getKeyValue( "TimezoneID" );
    }

    public void open( String sFileName_ ) {
        Util.debug( this, "open(..).sFileName_: " +
                    sFileName_ );

        // XML stuff
        try {

            ToDoTask.resetLastId();

            // turn the filename into an input source
            InputSource pInputSource = 
                Resolver.createInputSource( new File ( sFileName_ ) );

            // sax solution
            JimDocumentHandler handler = new JimDocumentHandler
                ( _getTimezoneID() );
            Parser parser = ParserFactory.makeParser( S_PARSER );
            parser.setDocumentHandler( handler );
            parser.setErrorHandler   ( handler );
            parser.parse( pInputSource );
            _pNotesPanel        .setNotes        ( handler.getNotes()         );
            _pToDoPanel         .setToDos        ( handler.getToDos()         );
            _pAnniversariesPanel.setAnniversaries( handler.getAnniversaries() );

        } catch ( SAXParseException err ) {
            System.out.println( "** Parsing error" 
                                + ", line " 
                                + err.getLineNumber ()
                                + ", uri " + err.getSystemId () );
            System.out.println("   " + err.getMessage ());
        } catch ( SAXException e ) {
            Util.printlnErr( this, "open(..).e: " + e );
            Exception   x = e.getException ();
                
            ((x == null) ? e : x).printStackTrace ();
        } catch ( Throwable t ) {
            Util.printlnErr( this, "open(..).t: " + t );
            t.printStackTrace ();
        }

        super.open( sFileName_ );
    }

    public class OpenAction extends RunnableAction {
        public OpenAction() {
            super( "&Open..." );
        }

        public void run() {
            open();
        }
    }

    protected String _getProjectSuffix() {
        String sProjectSuffix = getInit().getKeyValue( Init.getProjectSuffixKey() );
        if ( Util.isEmpty( sProjectSuffix ) ) {
            sProjectSuffix = ".jiml";
            getInit().setKeyValue( getInit().getProjectSuffixKey(), sProjectSuffix );
        }

        return sProjectSuffix;
    }

    /*public void save() {
        Util.debug( "jim.ViewController.save().ENTER" );
        String sTempFileName = getInit().getFileFullName();
        Util.debug( "ViewController.save().sTempFileName: " +
                    sTempFileName );

        if ( Util.isEmpty( sTempFileName ) ) {
            saveAs();

            return;
        }

        save( sTempFileName );
        }*/

    private ElementNode _createNOTEElement( XmlDocument pXmlDocument_, 
                                            DefaultMutableTreeNode pNode ) 
    {
        ElementNode enNote = (ElementNode)pXmlDocument_.createElement( "NOTE" );
        Note pNote = (Note)pNode.getUserObject();

        ElementNode enKey = (ElementNode)pXmlDocument_.createElement( "KEY" );
        enNote.appendChild( enKey );

        Text pText = pXmlDocument_.createTextNode( pNote.toString() );
        enKey.appendChild( pText );
        
        ElementNode enValue = (ElementNode)pXmlDocument_.createElement( "VALUE" );
        enNote.appendChild( enValue );

        pText = pXmlDocument_.createTextNode( pNote.getNote() );
        enValue.appendChild( pText );

        for( int child = 0; child < pNode.getChildCount(); child++ ) {
            DefaultMutableTreeNode tnChild = (DefaultMutableTreeNode)pNode.getChildAt( child );
            ElementNode enChild = _createNOTEElement( pXmlDocument_, 
                                                      tnChild );
            enNote.appendChild( enChild );
        }

        return enNote;
    }

    private ElementNode _createTODOProjectElement( XmlDocument pXmlDocument_, 
                                                   DefaultMutableTreeNode pNode ) 
    {
        ElementNode enProject = (ElementNode)pXmlDocument_.createElement( "PROJECT" );
        Vector pVector = (Vector)pNode.getUserObject();

        ElementNode enName = (ElementNode)pXmlDocument_.createElement( "NAME" );
        enProject.appendChild( enName );

        Text pText = pXmlDocument_.createTextNode( pVector.toString() );
        enName.appendChild( pText );

        Enumeration eToDoTasks = pVector.elements();
        while( eToDoTasks.hasMoreElements() ) {
            ToDoTask pToDoTask = (ToDoTask)eToDoTasks.nextElement();
            String sId       = Util.ltoa( pToDoTask.getId() );
            String sTitle    = pToDoTask.getTitle();
            String sComment  = pToDoTask.getComment();
            String sStart    = pToDoTask.getStart();
            String sClosed   = pToDoTask.getClosed();
            String sFinished = pToDoTask.getFinished();
            String sDeadline = pToDoTask.getDeadline();
            String sPriority = pToDoTask.getPriority().toString();

            ElementNode enToDo = (ElementNode)pXmlDocument_.createElement( "TODO" );
            enProject.appendChild( enToDo );

            ElementNode enId = (ElementNode)pXmlDocument_.createElement( "ID" );
            enToDo.appendChild( enId );

            pText = pXmlDocument_.createTextNode( sId );
            enId.appendChild( pText );

            ElementNode enTitle = (ElementNode)pXmlDocument_.createElement( "TITLE" );
            enToDo.appendChild( enTitle );

            pText = pXmlDocument_.createTextNode( sTitle );
            enTitle.appendChild( pText );

            ElementNode enComment = (ElementNode)pXmlDocument_.createElement( "COMMENT" );
            enToDo.appendChild( enComment );

            pText = pXmlDocument_.createTextNode( sComment );
            enComment.appendChild( pText );

            ElementNode enStart = (ElementNode)pXmlDocument_.createElement( "START" );
            enToDo.appendChild( enStart );

            pText = pXmlDocument_.createTextNode( sStart );
            enStart.appendChild( pText );

            ElementNode enClosed = (ElementNode)pXmlDocument_.createElement( "CLOSED" );
            enToDo.appendChild( enClosed );

            pText = pXmlDocument_.createTextNode( sClosed );
            enClosed.appendChild( pText );

            ElementNode enFinished = (ElementNode)pXmlDocument_.createElement( "FINISHED" );
            enToDo.appendChild( enFinished );

            pText = pXmlDocument_.createTextNode( sFinished );
            enFinished.appendChild( pText );

            ElementNode enDeadline = (ElementNode)pXmlDocument_.createElement( "DEADLINE" );
            enToDo.appendChild( enDeadline );

            pText = pXmlDocument_.createTextNode( sDeadline );
            enDeadline.appendChild( pText );

            Enumeration eDependencies = pToDoTask.getDependencies().elements();
            while( eDependencies.hasMoreElements() ) {
                String sDependsOn = (String)eDependencies.nextElement();

                ElementNode enDependsOn = (ElementNode)pXmlDocument_.createElement( "DEPENDSON" );
                enToDo.appendChild( enDependsOn );

                pText = pXmlDocument_.createTextNode( sDependsOn );
                enDependsOn.appendChild( pText );
            }

            Enumeration eBlocks = pToDoTask.getBlocks().elements();
            while( eBlocks.hasMoreElements() ) {
                String sBlocks = (String)eBlocks.nextElement();

                ElementNode enBlocks = (ElementNode)pXmlDocument_.createElement( "BLOCKS" );
                enToDo.appendChild( enBlocks );

                pText = pXmlDocument_.createTextNode( sBlocks );
                enBlocks.appendChild( pText );
            }

            ElementNode enPriority = (ElementNode)pXmlDocument_.createElement( "PRIORITY" );
            enToDo.appendChild( enPriority );

            pText = pXmlDocument_.createTextNode( sPriority );
            enPriority.appendChild( pText );
        }
        
        for( int child = 0; child < pNode.getChildCount(); child++ ) {
            DefaultMutableTreeNode tnChild = (DefaultMutableTreeNode)pNode.getChildAt( child );
            ElementNode enChild = _createTODOProjectElement( pXmlDocument_, 
                                                             tnChild );
            enProject.appendChild( enChild );
        }

        return enProject;
    }

    public void save( String sProjectFullFileName_ ) {
        _backup( sProjectFullFileName_ );

        /*try {
            DataOutputStream pDataOutputStream = new DataOutputStream
                   ( new FileOutputStream( sProjectFullFileName_ ) );
            
            // notes
            {
                // iterate through tree
                DefaultMutableTreeNode tnRoot = _pNotesPanel.getNotes();
                Enumeration eNotes = new TreeEnumeration( tnRoot );
                while( eNotes.hasMoreElements() ) {
                    TreeNode tnNext = (TreeNode)eNotes.nextElement();
                    Note pNote = (Note)((DefaultMutableTreeNode)tnNext).getUserObject();
                    String sKey = "";
                    while( tnNext != null ) {
                        sKey = tnNext.toString() + "/" + sKey;
                        tnNext = tnNext.getParent();
                    }
                    sKey += "/";
                    pDataOutputStream.writeUTF( sKey );
                    pDataOutputStream.writeUTF( pNote.getNote() );
                }
            }
            pDataOutputStream.writeUTF( "" );
            
            // Anniversaries
            Enumeration eAnniversaries = _pAnniversariesPanel.getAnniversaries().elements();
            while( eAnniversaries.hasMoreElements() ) {
                Anniversary pAnniversary = (Anniversary)eAnniversaries.nextElement();
                String sName = pAnniversary.getName();
                String sDate = pAnniversary.getDate();
                pDataOutputStream.writeUTF( sName );
                pDataOutputStream.writeUTF( sDate );
            }
            
            pDataOutputStream.close();
        } catch( Exception pException ) {
        }*/

        // now experiment with xml stuff
        try {
            XmlDocument pXmlDocument = new XmlDocument();
            Writer out = new OutputStreamWriter( new FileOutputStream( sProjectFullFileName_ ) );

            Text pText = null;
            
            ElementNode enData = (ElementNode)pXmlDocument.
                   createElement( "DATA" );
            pXmlDocument.appendChild( enData );
            
            // Version
            ElementNode enVersion = (ElementNode)pXmlDocument.
                   createElement( "VERSION" );
            Util.debug( this, "save(..).initVersion: " +
                        getInit().getVersion() );
            pText = pXmlDocument.createTextNode
                   ( Util.itoa( getInit().getVersion() ) );
            enVersion.appendChild( pText );
            enData.appendChild( enVersion );

            // Release
            ElementNode enRelease = (ElementNode)pXmlDocument.
                   createElement( "RELEASE" );
            Util.debug( this, "save(..).initRelease: " +
                        getInit().getRelease() );
            pText = pXmlDocument.createTextNode
                   ( Util.itoa( getInit().getRelease() ) );
            enRelease.appendChild( pText );
            enData.appendChild( enRelease );
            
            // notes
            {
                ElementNode enNotes = (ElementNode)pXmlDocument.
                       createElement( "NOTES" );
                enData.appendChild( enNotes );

                // iterate through tree
                DefaultMutableTreeNode tnRoot = 
                       _pNotesPanel.getNotes();
                ElementNode enNote = _createNOTEElement( pXmlDocument, tnRoot );
                enNotes.appendChild( enNote );
            }

            // ToDo
            {
                ElementNode enToDos = (ElementNode)pXmlDocument.
                       createElement( "TODOS" );
                enData.appendChild( enToDos );

                // iterate through tree
                DefaultMutableTreeNode tnRoot = 
                       _pToDoPanel.getToDos();
                ElementNode enToDoProject = _createTODOProjectElement( pXmlDocument, tnRoot );
                enToDos.appendChild( enToDoProject );
            }

            // anniversaries
            {
                ElementNode enAnniversaries = 
                       (ElementNode)pXmlDocument.
                       createElement( "ANNIVERSARIES" );
                enData.appendChild( enAnniversaries );

                Enumeration eAnniversaries = _pAnniversariesPanel.getAnniversaries().elements();
                while( eAnniversaries.hasMoreElements() ) {
                    Anniversary pAnniversary = (Anniversary)eAnniversaries.nextElement();
                    String sName = pAnniversary.getName();
                    String sDate = pAnniversary.getDate();

                    ElementNode enAnniversary = (ElementNode)pXmlDocument.createElement( "ANNIVERSARY" );
                    enAnniversaries.appendChild( enAnniversary );

                    ElementNode enName = (ElementNode)pXmlDocument.createElement( "NAME" );
                    enAnniversary.appendChild( enName );

                    pText = pXmlDocument.createTextNode( sName );
                    enName.appendChild( pText );
        
                    ElementNode enDate = (ElementNode)pXmlDocument.createElement( "DATE" );
                    enAnniversary.appendChild( enDate );

                    pText = pXmlDocument.createTextNode( sDate );
                    enDate.appendChild( pText );
                }
            }

            pXmlDocument.write( out );
            out.flush();
        } catch( Exception pException ) {
            Util.printlnErr( pException );
            SwingUtil.showMessage( this
                                   , "ERROR: Saving file\n'"
                                     + sProjectFullFileName_
                                     + "'\nfailed: " 
                                     + pException );
        }
    }

    private class SaveAction extends RunnableAction {
        public SaveAction() {
            super( "&Save" );
        }

        public void run() {
            save();
        }

        public KeyStroke getAccelerator() {
            return KeyStroke.getKeyStroke( KeyEvent.VK_S,
                                           Event.CTRL_MASK );
        }
    }

    /*public void saveAs() {
        // find file name
        String sProjectSuffix = _getProjectSuffix();
        String sOldProjectPath = getInit().getFilePath().replace('/', File.separatorChar);
        String sFileName = SwingUtil.getFileName
               ( this, 
                 "Save As",
                 "*" + sProjectSuffix,
                 sOldProjectPath,
                 "Save As" );
        if ( sFileName == null) {
            return;
        }

        // filename must end with ".jiml"
        if ( !Util.endsWith( sFileName, sProjectSuffix ) ) {
            SwingUtil.showMessage( ViewController.this, "Sorry, project file name must end with '" +
                                   sProjectSuffix + "' or '." +
                                   S_NEXTRECALL +"'!" );
            
            return;
        }
        if ( FileUtil.existsDir( sFileName ) ) {
            SwingUtil.showMessage( this, 
                                   "Sorry, file\n\n'" +
                                   sFileName +
                                   "'\n\nis a directory." +
                                   "Please choose another file name to save your data." );

            // try it again
            saveAs();
                
            return;
        }
        if ( FileUtil.existsFile( sFileName ) ) {
            boolean bOK = SwingUtil.isOKOrCancel( this, 
                                                  "File\n\n'" +
                                                  sFileName +
                                                  "'\n\nexists already. Is it ok to overwrite it?." );
            if ( !bOK ) {
                return;
            }
        }   
        getInit().setFileFullName( sFileName );

        save( sFileName );
        }*/

    private class SaveAsAction extends RunnableAction {
        public SaveAsAction() {
            super( "Save &As..." );
        }

        public void run() {
            ViewController.this.saveAs();
        }
    }

    private String _getNoteHTMLFileName( DefaultMutableTreeNode tnNote_ ) 
    {
        return _getHTMLFileName( tnNote_, "Notes" );
    }

    private String _getToDoHTMLFileName( DefaultMutableTreeNode tnToDo_ ) 
    {
        return _getHTMLFileName( tnToDo_, "ToDo" );
    }

    private String _getHTMLFileName( DefaultMutableTreeNode tnNode_, String sPrefix ) 
    {
        Object objNodes = tnNode_.getUserObject();

        String sFileName = sPrefix;
        String sProjectName = getInit().getFileName();
        sFileName += sProjectName;
        String sTail = objNodes.toString();
        DefaultMutableTreeNode tnParent = 
               (DefaultMutableTreeNode)tnNode_.getParent();
        while( tnParent != null ) {
            sTail = tnParent.getUserObject().toString() +
                   sTail;
            tnParent = (DefaultMutableTreeNode)tnParent.
                   getParent();
        }
        sFileName += Util.replace( sTail, " ", "" );
        sFileName += ".html";

        return _escapeFileName( sFileName );
    }

    private static final String S_NAME_SEPARATOR = " / ";

    private String _getHTMLName( DefaultMutableTreeNode tnToDo_ ) 
    {
        DefaultMutableTreeNode tnParent = 
               (DefaultMutableTreeNode)tnToDo_.getParent();
        String sToDoName = tnToDo_.getUserObject().toString();
        while( tnParent != null ) {
            sToDoName = tnParent.getUserObject().toString() + 
                   S_NAME_SEPARATOR +
                   sToDoName;
            tnParent = (DefaultMutableTreeNode)tnParent.
                   getParent();
        }

        return sToDoName;
    }

    private void _exportToDoProject( String sDirectory_,
                                     DefaultMutableTreeNode tnToDo_ ) 
    {
        String sProjectName = getInit().getFileName();

        String sHTML = "<html>\n<head>\n<title>" + sProjectName + 
               ": " + 
               _getHTMLName( tnToDo_ ) +
               "</title>\n</head>\n" +
               "<body text=\"#000000\" bgcolor=\"#ffffff\" link=\"#0000EF\" vlink=\"#51188E\" alink=\"#FF0000\">\n" +
               "<h1>" + sProjectName + ": ";

        //
        // headline
        //
        DefaultMutableTreeNode tnParent = null;
        {
            tnParent = (DefaultMutableTreeNode)tnToDo_.
                   getParent();
            if ( tnParent != null ) {
                sHTML += "<a href=\"";
                sHTML += _getToDoHTMLFileName( tnParent );
                sHTML += "\">";
                sHTML += _getHTMLName( tnParent );
                sHTML += "</a>" + S_NAME_SEPARATOR;
            }
            String sToDoName = tnToDo_.getUserObject().toString();
            
            sHTML += sToDoName + "</h1>\n";
        }

        //
        // date
        // 
        sHTML += _getHTMLDate();

        //
        // children
        //
        if ( tnToDo_.getChildCount() > 0 ) {
            sHTML += "<h2>Sub ToDo Lists</h2>\n";
            sHTML += "<ul>\n";
            
            for( int child = 0; 
                 child < tnToDo_.getChildCount();
                 child++ )
            {
                DefaultMutableTreeNode tnChild = 
                       (DefaultMutableTreeNode)tnToDo_.
                       getChildAt( child );

                // this is where the recursion takes place :-) !!
                _exportToDoProject( sDirectory_, tnChild );

                sHTML += "<li><a href=\"";
                sHTML += _getToDoHTMLFileName( tnChild );
                sHTML += "\">";
                
                sHTML += _getHTMLName( tnChild );

                sHTML += "</a></li>\n";
            }

            sHTML += "</ul>\n";
        }

        //
        // Tasks
        //
        {
            sHTML += "<h2>ToDo Tasks</h2>\n";
            sHTML += "<table border=\"0\" width=\"100%\">\n";
            
            sHTML += "<tr>\n";

            sHTML += "<td BGCOLOR=\"#000000\">";
            sHTML += "<font color=\"#FFFFFF\">";
            sHTML += "Nr.";
            sHTML += "</font>";
            sHTML += "</td>\n";

            sHTML += "<td BGCOLOR=\"#000000\">";
            sHTML += "<font color=\"#FFFFFF\">";
            sHTML += "Priority";
            sHTML += "</font>";
            sHTML += "</td>\n";

            sHTML += "<td BGCOLOR=\"#000000\">";
            sHTML += "<font color=\"#FFFFFF\">";
            sHTML += "Status";
            sHTML += "</td>\n";

            sHTML += "<td BGCOLOR=\"#000000\">";
            sHTML += "<font color=\"#FFFFFF\">";
            sHTML += "Date";
            sHTML += "</td>\n";

            sHTML += "<td BGCOLOR=\"#000000\">";
            sHTML += "<font color=\"#FFFFFF\">";
            sHTML += "Task";
            sHTML += "</td>\n";

            sHTML += "</tr>\n";

            Hashtable htPriorities = new Hashtable();
            htPriorities.put( ToDoTask.A, "Top"       );
            htPriorities.put( ToDoTask.B, "High"      );
            htPriorities.put( ToDoTask.C, "Normal"    );
            htPriorities.put( ToDoTask.D, "Low"       );

            Vector vToDos = (Vector)tnToDo_.getUserObject();
            Util.sortFast( vToDos, new ToDoPanel.ToDoComparable() );
            Enumeration eToDos = vToDos.elements();
            int open = 0;
            int closed = 0;
            while( eToDos.hasMoreElements() ) {
                sHTML += "<tr valign=TOP>\n";

                ToDoTask pToDoTask = (ToDoTask)eToDos.
                       nextElement();

                // Nr.
                int number = -1;
                if ( pToDoTask.isOpen() ) {
                    open++;
                    number = open;
                } else {
                    closed++;
                    number = closed;
                }
                sHTML += "<td align=RIGHT>" + number + "</td>\n";

                ToDoTask.Priority priority = pToDoTask.
                       getPriority();
                String sPriority = (String)htPriorities.get
                       ( priority );
                sHTML += "<td>" + sPriority + "</td>\n";

                sHTML += "<td>\n";
                sHTML += pToDoTask.getStatusHTML();
                sHTML += "</td>\n";

                sHTML += "<td>\n";
                sHTML += pToDoTask.getDate();
                sHTML += "</td>\n";

                sHTML += "<td><b>" + pToDoTask.getTitle() + "</b></td>\n";

                sHTML += "</tr>\n";

                // comment row
                String sCommentHTML = pToDoTask.getComment();
                sCommentHTML = Util.replace( sCommentHTML, "<", "&lt;" );
                sCommentHTML = Util.replace
                       ( pToDoTask.getComment(), "\n", "<br>" );
                if ( !Util.isEmpty( pToDoTask.getComment() ) ) {
                    sHTML += "<tr>\n";

                    for( int emptyCol = 0; emptyCol < 4; emptyCol++ ) {
                        sHTML += "<td>\n";
                        sHTML += "</td>\n";
                    }

                    sHTML += "<td>\n";
                    sHTML += sCommentHTML;
                    sHTML += "</td>\n";
                    
                    sHTML += "</tr>\n";
                }
            }

            sHTML += "</table>\n";
        }

        sHTML += "<hr>\n";
        sHTML += "<font size=-1>";
        sHTML += "Page generated by JIM (http://www.kclee.com/clemens/java/jim/).";
        sHTML += "</font>\n";
               
        sHTML += "</body>\n</html>\n";

        Util.debug( this, "_exportToDoProject(..).sHTML: " +
                    sHTML );

        String sFileName = _getToDoHTMLFileName( tnToDo_ );
        sFileName = _escapeFileName( sFileName );
        sFileName = FileUtil.concatPath( sDirectory_,
                                         sFileName );
        try {
            FileUtil.writeFile( sFileName, sHTML );
        } catch( Exception exception ) {
            SwingUtil.showMessage( this, "Error while writing file\n'" +
                                   sFileName +
                                   "'!" );
            if ( _bBatchMode )
            {
                Util.printlnErr( exception );
            }
        }
    }

    /**
     * No checks happen here anymore.
     */
    public void exportToDo( String sDirectory_ ) {
        DefaultMutableTreeNode tnToDosRoot = _pToDoPanel.
               getToDos();

        // walk through tree, might use a visitor in the future.
        _exportToDoProject( sDirectory_, tnToDosRoot );

        if ( !_bBatchMode )
        {
            SwingUtil.showMessage( this,
                                   "Finished export of all ToDo lists to directory:\n'" + sDirectory_ + "'" );
        }
    }

    private String _sLastExportDirectory = null;

    public void exportToDo() {
        // get directory
        if ( Util.isEmpty( _sLastExportDirectory ) ) {
            _sLastExportDirectory = (String)System.getProperties().
                   get( "user.home" );
        }
        String sDirectory = SwingUtil.getDirName
               ( this, 
                 "Export ToDo Directory",
                 _sLastExportDirectory );
        if ( Util.isEmpty( sDirectory ) ) {
            return;
        }
        
        if ( FileUtil.existsFile( sDirectory ) ) {
            SwingUtil.showMessage( this, "File\n'" +
                                   sDirectory +
                                   "'\nexists already.\n" +
                                   "Please specify a directory name next time. Amen." );

            return;
        }

        if ( !FileUtil.existsDir( sDirectory ) ) {
            boolean bCreateDir = SwingUtil.isOKOrCancel
                   ( this, "Directory\n'" +
                     sDirectory +
                     "'\ndoes not exist. Is it OK to create it now?" );
            if ( !bCreateDir ) {
                return;
            }
            
            FileUtil.md( sDirectory );
        }

        _sLastExportDirectory = sDirectory;

        exportToDo( sDirectory );
    }

    private void exportNote( String sDirectory_ ) {
        // so nothing gets lost what was typed!
        _pNotesPanel.saveLastNode();

        // get current note
        DefaultMutableTreeNode pNode = 
               _pNotesPanel.getSelectedNode();
        if ( pNode == null ) {
            return;
        }

        // get content
        String sNoteContent = ((Note)pNode.getUserObject()).
               getNote();

        //
        // construct file name
        //
        String sFileName = ((Note)pNode.getUserObject()).
               getName();
        
        while( pNode.getParent() != null ) {
            pNode = (DefaultMutableTreeNode)pNode.getParent();
            sFileName = ((Note)pNode.getUserObject()).
               getName() + sFileName;
        }
        sFileName = _escapeFileName( sFileName );
        sFileName = FileUtil.concatPath( sDirectory_,
                                         sFileName ) 
               + ".txt";

        //
        // save
        //
        try {
            FileUtil.writeFile( sFileName, sNoteContent );
            SwingUtil.showMessage( this,
                                   "Current Note saved as file\n'"
                                   + sFileName
                                   + "'!" );
        } catch( IOException exception ) {
            Util.printlnErr( "Exception writing file '"
                             + sFileName
                             + "': "
                             + exception );
            SwingUtil.showMessage( this,
                                   "Exception writing file\n'"
                                   + sFileName
                                   + "'!" );
        }
    }

    private void _exportAll( String sDirectory_ )
    {
        _exportNotes( sDirectory_ );
        exportToDo( sDirectory_ );
    }

    private void _exportNotes( String sDirectory_ ) 
    {
        // so nothing gets lost what was typed!
        _pNotesPanel.saveLastNode();

        // get current note
        DefaultMutableTreeNode pNode = 
               _pNotesPanel.getSelectedNode();

        _exportNoteToHTML( sDirectory_, pNode );

        if ( !_bBatchMode )
        {
            SwingUtil.showMessage( this,
                                   "Finished export of all Notes to directory:\n'" + sDirectory_ + "'" );
        }
    }

    private String _getHTMLDate()
    {
        Calendar calNow = Util.getCalendar
               ( getInit().getKeyValue( "TimezoneID" ) );
        String sRetVal = Util.getDate( calNow ) + "&nbsp;&nbsp;&nbsp;";
        sRetVal += Util.getTime( calNow ) + "<p>\n";

        return sRetVal;
    }

    /**
     * Saves also all sub notes!
     */
    private void _exportNoteToHTML( String sDirectory
                                   , DefaultMutableTreeNode pNode ) 
    {
        if ( pNode == null ) {
            return;
        }

        String sProjectName = getInit().getFileName();

        String sHTML = "<html>\n<head>\n<title>" + sProjectName + 
               ": " + 
               _getHTMLName( pNode ) +
               "</title>\n</head>\n" +
               "<body text=\"#000000\" bgcolor=\"#ffffff\" link=\"#0000EF\" vlink=\"#51188E\" alink=\"#FF0000\">\n" +
               "<h1>" + sProjectName + ": ";

        //
        // headline
        //
        DefaultMutableTreeNode tnParent = null;
        {
            tnParent = (DefaultMutableTreeNode)pNode.
                   getParent();
            if ( tnParent != null ) {
                sHTML += "<a href=\"";
                sHTML += _getNoteHTMLFileName( tnParent );
                sHTML += "\">";
                sHTML += _getHTMLName( tnParent );
                sHTML += "</a>" + S_NAME_SEPARATOR;
            }
            String sNoteName = pNode.getUserObject().toString();
            
            sHTML += sNoteName + "</h1>\n";
        }

        sHTML += _getHTMLDate();

        //
        // children
        //
        if ( pNode.getChildCount() > 0 ) {
            sHTML += "<h2>Sub Notes</h2>\n";
            sHTML += "<ul>\n";
            
            for( int child = 0; 
                 child < pNode.getChildCount();
                 child++ )
            {
                DefaultMutableTreeNode tnChild = 
                       (DefaultMutableTreeNode)pNode.
                       getChildAt( child );

                // this is where the recursion takes place :-) !!
                _exportNoteToHTML( sDirectory, tnChild );

                sHTML += "<li><a href=\"";
                sHTML += _getNoteHTMLFileName( tnChild );
                sHTML += "\">";
                
                sHTML += _getHTMLName( tnChild );

                sHTML += "</a></li>\n";
            }

            sHTML += "</ul>\n";
        }


        // get content
        sHTML += "\n<h2>Content:</h2>\n\n";
        sHTML += "<pre>\n";
        String sNoteContent = ((Note)pNode.getUserObject()).
               getNote();
        sHTML += sNoteContent + "</pre>\n";
        sHTML += "<hr>\n";
        sHTML += "<font size=-1>";
        sHTML += "Page generated by JIM (http://www.kclee.com/clemens/java/jim/).";
        sHTML += "</font>\n";
        sHTML += "</body>\n</html>\n";

        //
        // construct file name
        //
        String sFileName = FileUtil.concatPath( sDirectory
                                                , _getNoteHTMLFileName( pNode ) );

        //
        // save
        //
        try {
            FileUtil.writeFile( sFileName, sHTML );
            /*SwingUtil.showMessage( this,
                                   "Current Note saved as file\n'"
                                   + sFileName
                                   + "'!" );*/
        } catch( IOException exception ) {
            Util.printlnErr( "Exception writing file '"
                             + sFileName
                             + "': "
                             + exception );
            if ( !_bBatchMode )
            {
                SwingUtil.showMessage( this,
                                       "Exception writing file\n'"
                                       + sFileName
                                       + "'!" );
            }
        }
    }

    private String _sLastNoteExportDirectory = null;

    /**
     * @return   true on error if Jim should stop export.
     */
    private boolean _setNoteExportDirectory() 
    {
        // get directory
        if ( Util.isEmpty( _sLastNoteExportDirectory ) ) {
            _sLastNoteExportDirectory = getInit().getFilePath();
        }
        String sDirectory = SwingUtil.getDirName
               ( this, 
                 "Export Note Directory",
                 _sLastNoteExportDirectory );
        if ( Util.isEmpty( sDirectory ) ) {
            return true;
        }
        
        if ( FileUtil.existsFile( sDirectory ) ) {
            SwingUtil.showMessage( this, "File\n'" +
                                   sDirectory +
                                   "'\nexists already.\n" +
                                   "Please specify a directory name next time. Amen." );

            return true;
        }

        if ( !FileUtil.existsDir( sDirectory ) ) {
            boolean bCreateDir = SwingUtil.isOKOrCancel
                   ( this, "Directory\n'" +
                     sDirectory +
                     "'\ndoes not exist. Is it OK to create it now?" );
            if ( !bCreateDir ) {
                return true;
            }
            
            FileUtil.md( sDirectory );
        }

        _sLastNoteExportDirectory = sDirectory;

        return false;
    }

    private void exportNote() {
        if ( _setNoteExportDirectory() )
        {
            return;
        }

        exportNote( _sLastNoteExportDirectory );
    }

    private void _exportNotes() 
    {
        if ( _setNoteExportDirectory() )
        {
            return;
        }

        _exportNotes( _sLastNoteExportDirectory );
    }

    private class ExportToDoAction extends RunnableAction {
        public ExportToDoAction() {
            super( S_MF_EXPORT_TODO );
        }

        public void run() {
            exportToDo();
        }
    }

    private class ExportNoteAction extends RunnableAction {
        public ExportNoteAction() {
            super( S_MF_EXPORT_NOTE );
        }

        public void run() {
            exportNote();
        }
    }

    private class ExportNotesAction extends RunnableAction {
        public ExportNotesAction() {
            super( S_MF_EXPORT_NOTES );
        }

        public void run() {
            _exportNotes();
        }
    }

    private class ComponentViewerAction extends RunnableAction {
        public ComponentViewerAction() {
            super( "View Component Hierachy..." );
        }

        public void run() {
            ComponentHierachyFrame pComponentHierachyFrame =
                   new ComponentHierachyFrame();
            pComponentHierachyFrame.setVisible( true );
        }
    }

    private void setBoundsOrMaximize( Rectangle bounds_ ) {
        if ( bounds_ == null ) {
            SwingUtil.maximizeWindow( this );
        } else {
            setBounds( bounds_ );
        }
    }

    public class DateAction extends RunnableAction {
        public DateAction() {
            super( "Insert &Date" );
        }

        public void run() {
            // is notes pane visible?
            if ( _pNotesPanel.isVisible() ) {
                _pNotesPanel.insertDate();
            }
        }
    }

    private void _setMenuBar() {
	Vector vMenus = new Vector();
	Vector vFileMenu = new Vector();
	Vector vEditMenu = new Vector();
	Vector vHelpMenu = new Vector();

	vFileMenu.addElement( "File" );
	vFileMenu.addElement( new NewAction() );
	vFileMenu.addElement( new OpenAction() );
	vFileMenu.addElement( new SaveAction() );
	vFileMenu.addElement( new SaveAsAction() );
	vFileMenu.addElement( "---" );
	vFileMenu.addElement( new ExportNoteAction() );
	vFileMenu.addElement( new ExportNotesAction() );
	vFileMenu.addElement( new ExportToDoAction() );
	//vFileMenu.addElement( "---" );
	vFileMenu.addElement( "E&xit" );

	vEditMenu.addElement( "Edit" );
	vEditMenu.addElement( new DateAction() );

	vHelpMenu.addElement( "Help" );
	vHelpMenu.addElement( MainJMenuBar.S_HELP_CONTENTS );
	vHelpMenu.addElement( "-----" );
        vHelpMenu.addElement( new ComponentViewerAction() );
	vHelpMenu.addElement( "-----" );
	vHelpMenu.addElement( "&About..." );

	vMenus.addElement( vFileMenu );
	vMenus.addElement( vEditMenu );
	vMenus.addElement( vHelpMenu );

	setMenuBar( vMenus );
    }

    public ViewController() {
        super();
    }

    public ViewController( String[] asArg_, String sRCSHeader_ ) {
        super();

        initialize( asArg_, sRCSHeader_ );
    }

    private SplashWindow _pSplashWindow = null;

    private boolean _bInitialize = false;

    public void initialize( String[] asArg_, String sInfoHeader_ ) {
        if ( Util.isEmpty( sInfoHeader_ ) ) {
            sInfoHeader_ = Main.S_RCS_HEADER;
        }
        _bInitialize = true; // workaroudn for _confirmClose
	super.initialize( "Jim",
                          asArg_, sInfoHeader_ );
        _bInitialize = false;

	if ( isExitSet() ) {
	    return;
	}

        getInit().setAuthor( "Chr. Clemens Lee" );

        _bBatchMode = false;
        if ( getInit().getOptions().get( "exporthtml" ) != null )
        {
            _bBatchMode = true;
        }

	// start splash screen
        if ( _bBatchMode == false )
        {
            Image pImage = Toolkit.getDefaultToolkit().
                   getImage( this.getClass().
                             getResource( "jim_splash.gif" ) );
            _pSplashWindow = new SplashWindow
                   ( new Frame(), pImage );
            _pSplashWindow.setVisible( true );
            Util.debug( this, "initialize(..)._pSplashWindow: " +
                        _pSplashWindow );
        }

        _pNotesPanel = new NotesPanel( this );
        _pToDoPanel = new ToDoPanel( this );
        _pAnniversariesPanel = new AnniversariesPanel( this );

	setTitle();
	_setMenuBar();
        updateMenu();

        JTabbedPane pTabbedPane = new JTabbedPane();
        pTabbedPane.setName( "TabbedPane" );
        pTabbedPane.addTab( _pNotesPanel.getName(),
                            _pNotesPanel );
        pTabbedPane.addTab( _pToDoPanel.getName(),
                            _pToDoPanel );
        pTabbedPane.addTab( _pAnniversariesPanel.getName(),
                            _pAnniversariesPanel );

        pTabbedPane.addChangeListener( _pToDoPanel );

        getContentPane().setLayout( new BorderLayout() );
        getContentPane().add( pTabbedPane, BorderLayout.CENTER );

        String    sBounds = getInit().getKeyValue( "Bounds" );
        _bounds  = null;
        if ( !Util.isEmpty( sBounds ) ) {
            Vector vBounds = Util.stringToLines( sBounds, ',' );
            if ( vBounds.size() == 4 ) {
                int x = Util.atoi( (String)vBounds.elementAt( 0 ) );
                int y = Util.atoi( (String)vBounds.elementAt( 1 ) );
                int w = Util.atoi( (String)vBounds.elementAt( 2 ) );
                int h = Util.atoi( (String)vBounds.elementAt( 3 ) );
                Dimension dimScreen = SwingUtil.getScreenSize();
                if ( x < 0 ) x = 0;
                if ( y < 0 ) y = 0;
                if ( x + w > dimScreen.width  ) w = dimScreen.width  - x;
                if ( y + h > dimScreen.height ) h = dimScreen.height - y;
                _bounds = new Rectangle( x, y, w, h );
            }
        }

        if ( !_bBatchMode )
        {
            pack();
            setBoundsOrMaximize( _bounds );
        }

        Vector vArguments = getInit().getArguments();
        if ( vArguments.size() > 0 ) {
            String sFileToOpen = (String)vArguments.elementAt( 0 );
            Util.debug( this, "initialize(..).sFileToOpen: " +
                        sFileToOpen );
	    if ( !FileUtil.existsFile( sFileToOpen ) ) {
		SwingUtil.showMessage( this, "Sorry, the specified file\n\n'" +
				       sFileToOpen +
				       "'\n\ndoes not exist!" );
		newFile();
	    } else {
		open( sFileToOpen );
	    }
        } else if ( !Util.isEmpty( getInit().getFileFullName() ) ) {
            open( getInit().getFileFullName() );
        }

        if ( getInit().getOptions().get( "exporthtml" ) != null )
        {
            String sExportDir = getInit().getOptions().get( "exporthtml" ).toString();
            _exportAll( sExportDir );
        }

        if ( _bBatchMode == false )
        {
            show();
            setBoundsOrMaximize( _bounds );
            SwingUtilities.invokeLater( new Runnable() {
                    public void run() {
                        ViewController.this.setBoundsOrMaximize( _bounds );
                        Util.sleep( 1 );
                        ViewController.this.setBoundsOrMaximize( _bounds );
                        _pSplashWindow.dispose();
                        _pSplashWindow = null;
                    }
                } );
        }
    }

    public void addDateAction( JTextComponent text_ ) {
        TextAction pDateTextAction = new DateTextAction();
        Keymap kmOld = text_.getKeymap();
        Keymap kmNew = text_.addKeymap( null, kmOld );
        KeyStroke ksCtrlD = KeyStroke.getKeyStroke
               ( KeyEvent.VK_D, Event.CTRL_MASK );
        kmNew.removeKeyStrokeBinding( ksCtrlD );
        kmNew.addActionForKeyStroke
               ( ksCtrlD, pDateTextAction);
        text_.setKeymap( kmNew );
    }

    public class DateTextAction extends TextAction {
        public DateTextAction() {
            super( "DateTextAction" );
        }

        public void actionPerformed( ActionEvent pActionEvent_ ) {
            final JTextComponent txtTarget = getTextComponent( pActionEvent_ );

            if ( !txtTarget.isEnabled() ) {
                return;
            }
            SwingUtil.invokeLaterIfNecessary( new Runnable() {
                    public void run() {
                        String sDate = Util.getDate( Util.getCalendar() );
                        txtTarget.replaceSelection( sDate );
                    }
                } );
        }
    }

    public void start()
    {
        if ( _bBatchMode )
        {
            return;
        }

        super.start();
    }
}
